@@ -11,14 +11,13 @@ Authors:
 - Matthias Roels (implementation)
 """
 # std lib imports
-import json
-from typing import Optional
 import inspect
 from datetime import datetime
 import time
-
+import math
 import logging
-log = logging.getLogger(__name__)
+from random import shuffle
+
 # third party imports
 import pandas as pd
 from sklearn.model_selection import train_test_split
@@ -29,6 +28,8 @@ from cobra.preprocessing import KBinsDiscretizer
 from cobra.preprocessing import TargetEncoder
 from cobra.preprocessing import CategoricalDataProcessor
 
+log = logging.getLogger(__name__)
+
 
 class PreProcessor(BaseEstimator):
 
@@ -58,10 +59,7 @@ class PreProcessor(BaseEstimator):
     def __init__(self, categorical_data_processor: CategoricalDataProcessor,
                  discretizer: KBinsDiscretizer,
                  target_encoder: TargetEncoder,
-                 serialization_path: str=None,
-                 is_fitted: bool=False):
-
-        self.serialization_path = serialization_path
+                 is_fitted: bool = False):
 
         self._categorical_data_processor = categorical_data_processor
         self._discretizer = discretizer
@@ -71,23 +69,22 @@ class PreProcessor(BaseEstimator):
 
     @classmethod
     def from_params(cls,
-                    n_bins: int=10,
-                    strategy: str="quantile",
-                    closed: str="right",
-                    auto_adapt_bins: bool=False,
-                    starting_precision: int=0,
-                    label_format: str="{} - {}",
-                    change_endpoint_format: bool=False,
-                    regroup: bool=True,
-                    regroup_name: str="Other",
-                    keep_missing: bool=True,
-                    category_size_threshold: int=5,
-                    p_value_threshold: float=0.001,
-                    scale_contingency_table: bool=True,
-                    forced_categories: dict={},
-                    weight: float=0.0,
-                    imputation_strategy: str="mean",
-                    serialization_path: Optional[str]=None):
+                    n_bins: int = 10,
+                    strategy: str = "quantile",
+                    closed: str = "right",
+                    auto_adapt_bins: bool = False,
+                    starting_precision: int = 0,
+                    label_format: str = "{} - {}",
+                    change_endpoint_format: bool = False,
+                    regroup: bool = True,
+                    regroup_name: str = "Other",
+                    keep_missing: bool = True,
+                    category_size_threshold: int = 5,
+                    p_value_threshold: float = 0.001,
+                    scale_contingency_table: bool = True,
+                    forced_categories: dict = {},
+                    weight: float = 0.0,
+                    imputation_strategy: str = "mean"):
         """Constructor to instantiate PreProcessor from all the parameters
         that can be set in all its required (attribute) classes
         along with good default values.
@@ -144,8 +141,6 @@ class PreProcessor(BaseEstimator):
             Valid strategies are to replace with the global mean of the train
             set or the min (resp. max) incidence of the categories of that
             particular variable.
-        serialization_path : str, optional
-            path to save the pipeline to
 
         Returns
         -------
@@ -168,18 +163,17 @@ class PreProcessor(BaseEstimator):
 
         target_encoder = TargetEncoder(weight)
 
-        return cls(categorical_data_processor, discretizer, target_encoder,
-                   serialization_path)
+        return cls(categorical_data_processor, discretizer, target_encoder)
 
     @classmethod
-    def from_pipeline(cls, pipeline_path: str):
-        """Constructor to instantiate PreProcessor from a (fitted) pipeline,
-        stored as a JSON file.
+    def from_pipeline(cls, pipeline: dict):
+        """Constructor to instantiate PreProcessor from a (fitted) pipeline
+        which was stored as a JSON file and passed to this function as a dict.
 
         Parameters
         ----------
-        pipeline_path : str
-            path to the (fitted) pipeline
+        pipeline : dict
+            The (fitted) pipeline as a dictionary
 
         Returns
         -------
@@ -191,8 +185,6 @@ class PreProcessor(BaseEstimator):
         ValueError
             Description
         """
-        with open(pipeline_path, "r") as file:
-            pipeline = json.load(file)
 
         if not PreProcessor._is_valid_pipeline(pipeline):
             raise ValueError("Invalid pipeline")  # To do: specify error
@@ -235,6 +227,9 @@ class PreProcessor(BaseEstimator):
         log.info("Starting to fit pipeline")
         start = time.time()
 
+        # Ensure to operate on separate copy of data
+        train_data = train_data.copy()
+
         # Fit discretizer, categorical preprocessor & target encoder
         # Note that in order to fit target_encoder, we first have to transform
         # the data using the fitted discretizer & categorical_data_processor
@@ -264,11 +259,8 @@ class PreProcessor(BaseEstimator):
                  .format(time.time() - begin))
 
         self._is_fitted = True  # set fitted boolean to True
-        # serialize the pipeline to store the fitted output along with the
-        # various parameters that were used
-        self._serialize()
 
-        log.info("Fitting and serializing pipeline took {} seconds"
+        log.info("Fitting pipeline took {} seconds"
                  .format(time.time() - start))
 
     def transform(self, data: pd.DataFrame, continuous_vars: list,
@@ -351,107 +343,59 @@ class PreProcessor(BaseEstimator):
 
     @staticmethod
     def train_selection_validation_split(data: pd.DataFrame,
-                                         target_column_name: str,
-                                         train_prop: float=0.6,
-                                         selection_prop: float=0.2,
-                                         validation_prop: float=0.2,
-                                         stratify_split=True)->pd.DataFrame:
-        """Split dataset into train-selection-validation datasets and merge
-        them into one big DataFrame with an additional column "split"
-        indicating to which dataset the corresponding row belongs to.
+                                         train_prop: float = 0.6,
+                                         selection_prop: float = 0.2,
+                                         validation_prop: float = 0.2)-> pd.DataFrame:
+        """Adds `split` column with train/selection/validation values
+        to the dataset.
 
         Parameters
         ----------
         data : pd.DataFrame
             Input dataset to split into train-selection and validation sets
-        target_column_name : str
-            Name of the target column
         train_prop : float, optional
             Percentage data to put in train set
         selection_prop : float, optional
             Percentage data to put in selection set
         validation_prop : float, optional
             Percentage data to put in validation set
-        stratify_split : bool, optional
-            Whether or not to stratify the train-test split
 
         Returns
         -------
         pd.DataFrame
             DataFrame with additional split column
         """
-
-        if train_prop + selection_prop + validation_prop != 1.0:
+        if not math.isclose(train_prop + selection_prop + validation_prop, 1.0):
             raise ValueError("The sum of train_prop, selection_prop and "
-                             "validation_prop cannot differ from 1.0")
+                             "validation_prop must be 1.0.")
+
+        if train_prop == 0.0:
+            raise ValueError("train_prop cannot be zero!")
 
         if selection_prop == 0.0:
             raise ValueError("selection_prop cannot be zero!")
 
-        column_names = list(data.columns)
+        nrows = data.shape[0]
+        size_train = int(train_prop * nrows)
+        size_select = int(selection_prop * nrows)
+        size_valid = int(validation_prop * nrows)
+        correction = nrows - (size_train+size_select+size_valid)
 
-        predictors = [col for col in column_names if col != target_column_name]
+        split = ['train'] * size_train \
+                + ['train'] * correction \
+                + ['selection'] * size_select \
+                + ['validation'] * size_valid
 
-        # for the first split, take sum of selection & validation pct as
-        # test pct
-        test_prop = selection_prop + validation_prop
-        # To further split our test set into selection + validation set,
-        # we have to modify validation pct because we only have test_prop of
-        # the data available anymore for further splitting!
-        validation_prop_modif = validation_prop / test_prop
+        shuffle(split)
 
-        X = data[predictors]
-        y = data[target_column_name]
+        data['split'] = split
 
-        stratify = None
-        if stratify_split:
-            stratify = y
+        return data
 
-        X_train, X_test, y_train, y_test = train_test_split(
-            X, y,
-            test_size=test_prop,
-            random_state=42,
-            stratify=stratify
-            )
 
-        df_train = pd.DataFrame(X_train, columns=predictors)
-        df_train[target_column_name] = y_train
-        df_train["split"] = "train"
-
-        # If there is no validation percentage, return train-selection sets
-        # only
-        if validation_prop == 0.0:
-            df_selection = pd.DataFrame(X_test, columns=predictors)
-            df_selection[target_column_name] = y_test
-            df_selection["split"] = "selection"
-
-            return (pd.concat([df_train, df_selection])
-                    .reset_index(drop=True))
-
-        if stratify_split:
-            stratify = y_test
-
-        X_sel, X_val, y_sel, y_val = train_test_split(
-            X_test, y_test,
-            test_size=validation_prop_modif,
-            random_state=42,
-            stratify=stratify
-            )
-
-        df_selection = pd.DataFrame(X_sel, columns=predictors)
-        df_selection[target_column_name] = y_sel
-        df_selection["split"] = "selection"
-
-        df_validation = pd.DataFrame(X_val, columns=predictors)
-        df_validation[target_column_name] = y_val
-        df_validation["split"] = "validation"
-
-        return (pd.concat([df_train, df_selection, df_validation])
-                .reset_index(drop=True))
-
-    def _serialize(self) -> dict:
+    def serialize_pipeline(self) -> dict:
         """Serialize the preprocessing pipeline by writing all its required
-        parameters to a JSON file.
+        parameters to a dictionary to later store it as a JSON file.
 
         Returns
         -------
@@ -474,14 +418,6 @@ class PreProcessor(BaseEstimator):
 
         pipeline["_is_fitted"] = True
 
-        if self.serialization_path:
-            path = self.serialization_path
-        else:
-            path = "./pipeline_tmp.json"
-
-        with open(path, "w") as file:
-            json.dump(pipeline, file)
-
         return pipeline
 
     @staticmethod
